@@ -60,6 +60,8 @@
 #include "ngtcp2_pcg_test.h"
 #include "ngtcp2_ratelim_test.h"
 #include "ngtcp2_conn_info_test.h"
+#include "ngtcp2_cid_test.h"
+#include "ngtcp2_log_test.h"
 
 int main(int argc, char *argv[]) {
   const MunitSuite suites[] = {
@@ -93,6 +95,8 @@ int main(int argc, char *argv[]) {
     pcg_suite,
     ratelim_suite,
     conn_info_suite,
+    cid_suite,
+    log_suite,
     {0},
   };
   const MunitSuite suite = {
