@@ -50,6 +50,8 @@ munit_void_test_decl(test_ngtcp2_conn_recv_reset_stream)
 munit_void_test_decl(test_ngtcp2_conn_recv_stop_sending)
 munit_void_test_decl(test_ngtcp2_conn_recv_stream_data_blocked)
 munit_void_test_decl(test_ngtcp2_conn_recv_data_blocked)
+munit_void_test_decl(test_ngtcp2_conn_recv_streams_blocked)
+munit_void_test_decl(test_ngtcp2_conn_recv_new_token)
 munit_void_test_decl(test_ngtcp2_conn_recv_conn_id_omitted)
 munit_void_test_decl(test_ngtcp2_conn_short_pkt_type)
 munit_void_test_decl(test_ngtcp2_conn_recv_stateless_reset)
