@@ -40,5 +40,7 @@ munit_void_test_decl(test_ngtcp2_map)
 munit_void_test_decl(test_ngtcp2_map_functional)
 munit_void_test_decl(test_ngtcp2_map_each)
 munit_void_test_decl(test_ngtcp2_map_clear)
+munit_void_test_decl(test_ngtcp2_map_free)
+munit_void_test_decl(test_ngtcp2_map_remove)
 
 #endif /* !defined(NGTCP2_MAP_TEST_H) */
